#define LINT_ARGS 1      /* Enable Compiler Parameter Checking */
#include <stdio.h>
#include <math.h>

/*

Estimate means and covariances from a set of sample data.

The input data is divided into "groups" of observables.  
Within each group a full covariance matrix is accumulated.  
Currently all groups are the same size.

The input format is:

     1) A parameter header largely irrelevant to this program,
        except that two integers - 'num_grps' and 'grp_size', 
        defining the grouping of input data - are extracted
        from it.

     2) A series, terminated by EOF, of "configurations".  Each 
        configuration is a set of (num_grps * grp_size) floating-
        point numbers.

*/


/* Data sizes... */

#define MAX_MEANS 1000
#define MAX_COVNS 250000
#define MAX_GRPS  1000


/* Static workspace... */

double means_acc [MAX_MEANS] ;
double covns_acc [MAX_COVNS] ;

double *covn_ptrs [MAX_MEANS] ;



int init_arrays (mean_grp, covn_grp, num_grps, grp_size)

double  *mean_grp [MAX_GRPS] ;
double **covn_grp [MAX_GRPS] ;
int num_grps, grp_size ;

/*

  Initialise pointer arrays so that 'mean_grp' looks like an array of 
  linear arrays and 'covn_grp' looks like and array of triangular arrays.

  Also initialise data to zeros.

*/

{
  {
    int num_means = num_grps * grp_size ;

    if (num_grps > MAX_GRPS)
    {
      printf ("\nERRROR: Too many groups of observables\n") ;
      return 0 ;
    }
    if (num_means > MAX_MEANS)
    {
      printf ("\nERRROR: Too many observables\n") ;
      return 0 ;
    }
    if (((num_means * (num_means + 1)) / 2) > MAX_COVNS)
    {
      printf ("\nERROR: Too many covariances\n") ;
      return 0 ;
    }
  }

  {
    /*

    Initialise pointer arrays...

    double  *mean_grp [MAX_GRPS] ;
    double **covn_grp [MAX_GRPS] ;

    and (static)

    double *covn_ptrs [MAX_MEANS] ;

    */

    int g, i ;
    double *mean_addr = means_acc ;
    double *covn_addr = covns_acc ;
    double **covn_ptr_addr = covn_ptrs ;

    for (g = 0; g < num_grps; g++)
    {
      mean_grp [g] = mean_addr ;
      covn_grp [g] = covn_ptr_addr ;
      for (i = 0; i < grp_size; i++)
      {
        covn_grp [g] [i] = covn_addr ;
        mean_addr     ++ ;
        covn_ptr_addr ++ ;
        covn_addr     += i + 1 ;
      }
    }
  }

  {

    /* Finally, zero arrays */

    int g, i, j ;

    for (g = 0; g < num_grps; g++)
      for (i = 0; i < grp_size; i++)
      {
        mean_grp [g] [i] = 0.0 ;
        for (j = 0; j <= i; j++)
          covn_grp [g] [i] [j] = 0.0 ;
      }
  }

  return 1 ;
}

int add_sample (mean_grp, covn_grp, num_grps, grp_size)

double  *mean_grp [MAX_GRPS] ;
double **covn_grp [MAX_GRPS] ;
int num_grps, grp_size ;

/*

   Get a configuration, and add its contribution to arrays.

*/

{
  int g, i, j ;

  for (g = 0; g < num_grps; g++)
  {
    float obs [MAX_MEANS] ;
    {
      /* Read in group data */

      for (i = 0; i < grp_size; i++)
      {
        int result ;

        result = scanf ("%f", obs + i) ;
        if (result == EOF)
          if (g | i)
          {
            printf ("\nERROR: Unexpected end of file\n") ;
            exit (0) ;
          }
          else
            return 0 ;
      }

      /* Add it to arrays */

      for (i = 0; i < grp_size; i++)
      {
        mean_grp [g] [i] += obs [i] ;
        for (j = 0; j <= i; j++)
          covn_grp [g] [i] [j] += obs [i] * obs [j] ;
      }
    }
  }
  return 1 ;
}

void output_group (means, covns, grp_size) 

double  *means ;
double **covns ;
int grp_size ;

/*

  Pretty-print group of means and covariances.

*/

{
  if (grp_size == 1)
    printf ("%13e %13e\n", means [0], covns [0] [0]) ;
  else
  {
    int i, j, k ;

    for (i = 0; i < grp_size; i++)
      printf ("%13e\n", means [i]) ;

    printf ("\n") ;

    k = 0 ;
    for (i = 0; i < grp_size; i++)
      for (j = 0; j <= i; j++)
      {
        printf ("%13e ", covns [i] [j]) ;
        k++ ;
        if (k == 5)
        {
          printf ("\n") ;
          k = 0 ;
        }
      }

    printf ("\n\n") ;
  }
}

void process_sums (mean_grp, covn_grp, num_grps, grp_size, num_samples) 

double  *mean_grp [MAX_GRPS] ;
double **covn_grp [MAX_GRPS] ;
int num_grps, grp_size, num_samples ;

/*

  Subtractions, normalisations, and stuff.  Ouput results.

*/

{
  if (num_samples > 0)
  {
    int g, i, j ;

    for (g = 0; g < num_grps; g++)
    {

      /* Normalise means */

      for (i = 0; i < grp_size; i++)
        mean_grp [g] [i] /= num_samples ;

      /* Subtractions from covariances */

      for (i = 0; i < grp_size; i++)
        for (j = 0; j <= i; j++)
          covn_grp [g] [i] [j] = (covn_grp [g] [i] [j] / num_samples) -
                                 (mean_grp [g] [i] * mean_grp [g] [j]) ;

      /* Estimate parent distribution covariances */

      for (i = 0; i < grp_size; i++)
        for (j = 0; j <= i; j++)
          if (num_samples > 1)
          {
            covn_grp [g] [i] [j] /= num_samples - 1 ;
          }
          else
            covn_grp [g] [i] [j] = 0.0 ;

      /* Output results */

      output_group (mean_grp [g], covn_grp [g], grp_size) ;
    }
  }
  else
    printf ("\nERROR: No data!\n") ;
}

void main ()
{
  int num_grps, grp_size ;
  double  *mean_grp [MAX_GRPS] ;
  double **covn_grp [MAX_GRPS] ;

  read_header (&num_grps, &grp_size) ;

  if (init_arrays (mean_grp, covn_grp, num_grps, grp_size))
  {
    int num_samples = 0 ;

    while (add_sample (mean_grp, covn_grp, num_grps, grp_size))
      num_samples ++ ;

    printf ("%d\n\n", num_samples) ;
    process_sums (mean_grp, covn_grp, num_grps, grp_size, num_samples) ;
  }
}

/*
--
--     End of File.
--
*/
